#######################################################################
# Welcome to PeacoQC - CYTO 2022
######################################################################

## This will force the creation of a backup input and output CSV file for additional debugging
debug_input_output <- FALSE
## For debugging to see what FlowJo is passing to the script.
##
## cat("INPUT-IT                = INPUT_IT \n")
## cat("INPUT-MD          		= INPUT_MAD \n")
## cat("FJ-PAR-Cytof            = FJ_PAR_Cytof \n")
## cat("FJ-PAR-QCSAMPLES        = FJ_PAR_QCSAMPLES \n")
## cat("OUT-PATH            	= OUT_PATH \n")
## cat("FJ-DATA-FILE-PATH"       = FJ_DATA_FILE_PATH \n")
## cat("FJ-PARAMS-LIST          = FJ_PARAMS_LIST \n")



## INSTALL REQUIRED PACKAGES ##
#######################
# Install function for packages

loadPackages<-function(...){
	libs<-unlist(list(...))
	req<-unlist(lapply(libs,require,character.only=TRUE))
	need<-libs[req==FALSE]
	if(length(need)>0){
		tryCatch({
			install.packages(pkgs=need, repos="https://cloud.r-project.org")
		}, error = function(e){
			tryCatch({
				install.packages(pkgs=need, repos="https://cloud.r-project.org", type = 'binary')
			}, error = function(e){
				dir.create(path = Sys.getenv("R_LIBS_USER"), showWarnings = FALSE, recursive = TRUE)
				install.packages(pkgs=need, lib = Sys.getenv("R_LIBS_USER"), repos="https://cloud.r-project.org", type = 'binary')
			})
		})
		lapply(need,require,character.only=TRUE)
	}
}


loadBioc<-function(...){
	libs<-unlist(list(...))
	req<-unlist(lapply(libs,require,character.only=TRUE))
	need<-libs[req==FALSE]
	if(length(need)>0){
		tryCatch({
			BiocManager::install(need)
		}, error = function(e){
			tryCatch({
				BiocManager::install(need, type = 'binary')
			}, error = function(e){
				BiocManager::install(need, lib = Sys.getenv("R_LIBS_USER"), type = 'binary')
			})
		})
		lapply(need,require,character.only=TRUE)
	}
}

# Try to write a special function for devtools installs:
devPackages<-function(x){
	devtools::install_github(x);
}

# the Real Deal (aka required packages) #
loadPackages(c('BiocManager', 'devtools', 'flowCore'))

if (!require("PeacoQC",character.only=TRUE)){
	tryCatch({
		devPackages('saeyslab/PeacoQC')
	}, error = function(e){
		loadBioc('PeacoQC')
	})
}else{
	require('PeacoQC',character.only=TRUE)
}

loadBioc('ComplexHeatmap');


######################
######################
## Last tested with R 4.2.0 (2022-06-05)
## Check what the user has and compare to what we know works
sessionInfo()

#Inputs from Java
inputIT = INPUT_IT
inputMAD = INPUT_MAD
cytof = FJ_PAR_Cytof
qcSamples = TRUE #FJ_PAR_QCSAMPLES
OutputPath = "FJ_OUPUT_FILE"
FolderPath = "FJ_DATA_FILE_PATH"
# Need to add a parameter selector.
# parNames <- "FJ_PARAMS_LIST"

robust.read.FCS <- function(fcsPath) {
	out <- tryCatch(
	{
		read.FCS(fcsPath, dataset=dataset)
	},
	error=function(cond) {
		read.FCS(fcsPath, dataset=dataset, emptyValue=FALSE, ignore.text.offset=TRUE)
	})
	timeIndex <- -1
	timeIndex <- tryCatch(
	{
		match("time", tolower(colnames(out)))
	},
	error=function(cond) {
		-1
	})
	if (!(is.na(timeIndex)) && (timeIndex > 0) && (min(exprs(out)[,timeIndex]) == max(exprs(out)[,timeIndex]))) {
		## Our time has a single value, that may be because there is a wrong $PnR, so let's make flowCore ignore it
		out <- tryCatch(
		{
			read.FCS(fcsPath, dataset=dataset, truncate_max_range = FALSE)
		},
		error=function(cond) {
			read.FCS(fcsPath, dataset=dataset, truncate_max_range = FALSE, emptyValue=FALSE, ignore.text.offset=TRUE)
		})
	}

	return(out)
}


####### Based on PeacoQC 1.7.0

## remove last slash if present
strip.sep <- function(name) {
	ifelse(substr(name,nchar(name),nchar(name))==.Platform$file,
	substr(name,1,nchar(name)-1),name)
}


all_data<-lapply(list.files(FolderPath, full.names=T, pattern = "\\.fcs$"), function(x)read.FCS(x,transformation=F));
# inputFCS <- robust.read.FCS("FJ_DATA_FILE_PATH");



# eventsCount <- length(inputFCS@exprs[,1])
FCSParams <- c(all_data[[1]]@parameters@data[["name"]])


# # Compensate and transform the data
tryCatch({
	all_data <- lapply(all_data, function(x)flowCore::compensate(x, flowCore::keyword(x)$SPILL))
}, error = function(e){
	tryCatch({
		all_data <- lapply(all_data, function(x)flowCore::compensate(x, flowCore::keyword(x)$`$SPILLOVER`))
	} , error = function(e){
		print('Could not find compensation.')
	}
	)
}
)

tryCatch({
	all_data <- lapply(all_data, function(x)flowCore::transform(x, flowCore::estimateLogicle(x, colnames(flowCore::keyword(x)$SPILL))))
}, error = function(e){
tryCatch({
	all_data <- lapply(all_data, function(x)flowCore::transform(x, flowCore::estimateLogicle(x, colnames(flowCore::keyword(x)$`$SPILLOVER`))))
}, error = function(e){
	all_data <- lapply(all_data, function(x)flowCore::transform(x, flowCore::estimateLogicle(x, FCSParams)))
}
)
}
)




if (!("time" %in% tolower(FCSParams))){
stop("The input FCS file does not contain the TIME parameter.", call.=FALSE)
}


# Remove the Margins  - This removes too much right now.
# all_data <-lapply(all_data, function(x)RemoveMargins(x, channels=FCSParams, output="frame"));



cleanUpAfterwards <- TRUE #FJ_PAR_DELETE_WD
originalDir <- getwd()



tryCatch({pdf(NULL)}, error = function(e){})

if (cytof == TRUE){
	newFlowFrame <- tryCatch({
		all_data_clean <- lapply(all_data, function(x)PeacoQC(x,
		channels= 1:length(FCSParams),
		determine_good_cells="all",
		save_fcs = TRUE,
		plot=TRUE,
		IT_limit = inputIT,
		MAD = inputMAD,
		output_directory = "FJ_DATA_FILE_PATH"));
	}, error = function(e){
		all_data_clean <- lapply(all_data, function(x)PeacoQC(x,
		channels= 1:length(FCSParams),
		determine_good_cells="all",
		save_fcs = TRUE,
		plot=FALSE,
		IT_limit = inputIT,
		MAD = inputMAD,
		output_directory = "FJ_DATA_FILE_PATH"));
	})
} else {
	newFlowFrame <- tryCatch({
		all_data_clean <- lapply(all_data, function(x)PeacoQC(x,
		channels= 1:length(FCSParams),  #parNames,
		determine_good_cells="all",
		save_fcs = TRUE,
		plot=TRUE,
		IT_limit = inputIT,
		MAD = inputMAD,
		output_directory = OutputPath));
	}, error = function(e){
		all_data_clean <- lapply(all_data, function(x)PeacoQC(x,
		channels= 1:length(FCSParams),  #parNames,
		determine_good_cells="all",
		save_fcs = TRUE,
		plot=FALSE,
		IT_limit = inputIT,
		MAD = inputMAD,
		output_directory = OutputPath));
	})
}


# Make heatmap overview of the quality control run. I am not sure if these heatmaps are worth it. They are meant to be used over multiple samples.
# heatmapIMG <- PeacoQCHeatmap(report_location = paste(OutputPath,"/", fcsName,"/PeacoQC_results/PeacoQC_report.txt", sep = ""))


# R seems to be saving the .RData when exiting, so let's clean up to at least make that tiny (i.e., empty environment)
rm(list=ls())
